(* fichier compte-classe.ml : comptage et classement simultans *)
open Num

type cellule = {
  mutable a:int; b:int; c:int;   (* exposants de 2,3,5 *)
  mutable tiers     : cellule;   (* lien vers u/3 *)
  mutable triple    : cellule;   (* lien vers 3u  *)
  mutable cinquime : cellule;   (* lien vers u/5 *)
  mutable quintuple : cellule;   (* lien vers 5u  *)
  mutable prcdent : cellule;   (* lien arrire dans la suite 2-3-5 *)
  mutable suivant   : cellule;   (* lien avant dans la suite 2-3-5   *)
  mutable suiv35    : cellule    (* lien avant dans la suite 3-5     *)
}
let rec nil =
  {a=0;b=0;c=0;
   tiers     = nil; triple    = nil;
   cinquime = nil; quintuple = nil;
   prcdent = nil; suivant   = nil;
   suiv35    = nil}

              (* +---------------------------------------------+
                 |  Gnration des entiers de Hamming impairs  |
                 +---------------------------------------------+ *)

let troisfois  x = {x with b=x.b+1; suiv35=x}
and cinqfois   x = {x with c=x.c+1; suiv35=x}

let rec dbut =
  {a=0;b=0;c=0;
   tiers     = nil;   triple    = nil;
   cinquime = nil;   quintuple = nil;
   prcdent = dbut; suivant   = dbut;
   suiv35    = nil}


let h = ref(dbut)        (* dernier lment engendr *)
and y = ref(troisfois dbut)    (* prochain triple    *)
and z = ref(cinqfois  dbut)    (* prochain quintuple *)

(* calcul de minimum *)
let dmin y z = match (z.b>0, y.c>0) with
  |true,_    -> y
  |_,   true -> z
  |_,   _    -> if Int(3)**/Int(y.b) */ Int(5)**/Int(y.c) >/
                   Int(3)**/Int(z.b) */ Int(5)**/Int(z.c)
                then z else y

(* prolonge la liste de Hamming 3-5 d'un lment *)
let avance35() =
  let t = dmin !y !z in
  !h.suiv35 <- t;
  h := t;
  if t.b = !y.b & t.c = !y.c then begin
    let u = !y.suiv35 in
    t.tiers  <- u;
    u.triple <- t;
    y := troisfois u.suiv35
  end
  else t.tiers <- nil;
  if t.b = !z.b & t.c = !z.c then begin
    let u = !z.suiv35 in
    t.cinquime <- u;
    u.quintuple <- t;
    z := cinqfois u.suiv35
  end
  else t.cinquime <- nil

                  (* +--------------------------+
                     |  Classement et comptage  |
                     +--------------------------+ *)

let p  = ref(dbut) (* pointeur d'insertion dans la tranche *)
and a  = ref(-1)    (* dcalage exposant de 2 *)
and c1 = ref(0)     (* cardinal des tranches prcdentes *)
and c2 = ref(1)     (* cardinal de la tranche courante   *)

(* comparaison exacte *)
let diff x y = if x>y then (x-y,0) else (0,y-x)
let (<<) u v =
  let (a,a') = diff u.a v.a
  and (b,b') = diff u.b v.b
  and (c,c') = diff u.c v.c in
  Int(2)**/Int(a)  */ Int(3)**/Int(b)  */ Int(5)**/Int(c) </ 
  Int(2)**/Int(a') */ Int(3)**/Int(b') */ Int(5)**/Int(c')

(* insre un lment aprs la cellule pointe par q *)
let rec place q =
  !h.suivant <- q.suivant;
  q.suivant.prcdent <- !h;
  q.suivant <- !h;
  !h.prcdent <- q;
  p := !h;
  incr c2;
  avance35();
  avance235()

(* augmente la tranche courante d'un lment tant que possible *)
(* lve l'exception Exit quand la tranche est pleine           *)
and avance235() =

  !h.a <- - !a; (* mise  niveau exposant de 2 *)

  (* cherche un minorant  partir du tiers *)
  let min3 =
    let t = !h.tiers    in
    let u = t.prcdent in
    let v = u.triple    in
    if t == nil or u == dbut then nil
    else if u.a = t.a + v.a + !a then v
    else if u.a > t.a + v.a + !a then raise Exit
    else nil in
  
  (* cherche un minorant  partir du cinquime *)
  let min5 =
    let t = !h.cinquime in
    let u = t.prcdent  in
    let v = u.quintuple  in
    if t == nil or u == dbut then nil
    else if u.a = t.a + v.a + !a then v
    else if u.a > t.a + v.a + !a then raise Exit
    else nil in

  (* cherche un majorant  partir du tiers *)
  let max3 =
    let t = !h.tiers  in
    let u = t.suivant in
    let v = u.triple  in
    if t == nil or t == dbut then nil
    else if u.a = t.a + v.a + !a then v
    else if u.a > t.a + v.a + !a
    then if dbut << !h then raise Exit else dbut
    else nil in

  (* cherche un majorant  partir du cinquime *)
  let max5 =
    let t = !h.cinquime in
    let u = t.suivant    in
    let v = u.quintuple  in
    if t == nil or t == dbut then nil
    else if u.a = t.a + v.a + !a then v
    else if max3 == nil & u.a > t.a + v.a + !a
    then if dbut << !h then raise Exit else dbut
    else nil in

  (* si un majorant suit un minorant c'est bon *)
       if min3 != nil & (min3 == max3.prcdent or min3 == max5.prcdent) then place min3
  else if min5 != nil & (min5 == max3.prcdent or min5 == max5.prcdent) then place min5

  (* sinon recule  partir d'un majorant s'il y en a un *)
  else let q = ref(if max3 != nil then max3 else max5) in
    if !q != nil then begin
      while !q.prcdent != !p & !q.prcdent != min3 & !q.prcdent != min5 & !h << !q.prcdent
      do q := !q.prcdent done;
      place !q.prcdent
    end

    (* dernier recours: avance  partir d'un minorant ou de p *)
    else begin
      if min5 != nil then p := min5;
      if min3 != nil then p := min3;
      while !p.suivant != max3 & !p.suivant != max5 & !p.suivant << !h do
        p := !p.suivant;
        if !p == dbut then raise Exit
      done;
      place !p
    end
(* fin de avance235 (si, si) *)

(* lance les calculs *)
let _ =
  let n = int_of_string(Sys.argv.(1)) in
  avance35();
  while !c1 <= n do
    incr a;
    p := dbut;
    (try avance235() with Exit -> ());
    c1 := !c1 + !c2;
  done;
  let x = ref(dbut) in
  for i = !c1-1 downto n+1 do x := !x.prcdent done;
  Printf.printf "2^%d . 3^%d * 5^%d" (!x.a + !a) !x.b !x.c;
  print_newline()
