(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                  Calcul du n-me nombre de Hamming                       |
  |                                                                          |
  |                       Mthode volumtrique                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

(* Michel Quercia, le 08/06/99 *)
(* compilation :
   ocamlopt -o hamming nums.cmxa unix.cmxa volume.ml -cclib -lnums -cclib -lunix
*)

                         (*+-------------------------+
                           |  Messages de dboguage  |
                           +-------------------------+*)

let debug = ref false   (* option -debug *)
open Printf
open Unix

let t0 = ref(0.0) (* dernier relev chrono *)

let message texte = if !debug then begin
  let t1 = (let t = times() in t.tms_utime +. t.tms_stime) in
  fprintf Pervasives.stderr "%5.2f %5.2f %s" t1 (t1 -. !t0) (texte());
  t0 := t1;
  flush Pervasives.stderr
end

let mesg texte = if !debug then begin
  prerr_string(texte());
  flush Pervasives.stderr
end



(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                             Descriptions abstraites                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

                           (* +-------------------+
                              |  nombres entiers  |
                              +-------------------+ *)

module type Entiers = sig 
  type t

  val ( ++ ) : t -> t -> t                  (* oprations usuelles *)
  val ( -- ) : t -> t -> t
  val ( ** ) : t -> t -> t
  val ( // ) : t -> t -> t
  val (<<= ) : t -> t -> bool
  val ( == ) : t -> t -> bool

  val zro   : t                 (* on aura besoin de tous ceux l *)
  val un     : t
  val deux   : t
  val trois  : t
  val six    : t
  val douze  : t

  val of_string : string -> t           (* fonctions de conversion *)
  val string_of : t -> string

  val cubroot   : t -> t           (* racine cubique approximative *)
end

                          (* +---------------------+
                             |  Calcul de surface  |
                             +---------------------+ *)

module type Surf = functor (Z:Entiers) -> sig
  open Z
  type coef                                (* coefficients a,b     *)
  val  make    : t * t -> coef             (* enregistre a,b       *)
  val  surface : coef -> t -> t            (* surface ax+by <= n   *)

end

                          (* +--------------------+
                             |  Calcul de volume  |
                             +--------------------+ *)

module type Vol = functor (Z:Entiers) -> sig
  open Z
  type coef                          (* coefficients a,b,c, hmax et pmax *)
  val  make    : t * t * t * t * int -> coef    (* enregistre a,b,c,h,p  *)
  val  volume  : coef -> t -> t                 (* volume ax+by+cz <= n  *)
end

            (* +-------------------------------------------------+
               |  Recherche d'un entier de Hamming ou d'un rang  |
               +-------------------------------------------------+ *)

module type Ham = functor (Z:Entiers) -> sig
  open Z
  val hamming : t -> int -> t -> t -> t -> t -> unit
  val rang    : t -> int -> t -> t -> t -> t -> t -> t -> unit

  (* hamming   hmax  pmax   a    b    c    k           *)
  (* rang      hmax  pmax   a    b    c    x    y    z *)
end

(* +------------------------------------------------------------------------+
   |                                                                        |
   |                           Implmentation des entiers                   |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* entiers de longueur arbitraire *)
module Znum = struct
  open Num
  type t = num
  let ( ++ )    = ( +/ )
  and ( -- )    = ( -/ )
  and ( ** )    = ( */ )
  and ( // )    = quo_num
  and (<<= )    = (<=/ )
  and ( == )    = ( =/ )
  and zro      = Int(0)
  and un        = Int(1)
  and deux      = Int(2)
  and trois     = Int(3)
  and six       = Int(6)
  and douze     = Int(12)
  and of_string = num_of_string
  and string_of = string_of_num
  and cubroot x = let y = floor(float_of_num(x)**(1.0/.3.0)) in
                  num_of_string(Printf.sprintf "%0.0f" y)
end

(* flottants = entiers de 64 bits *)
module Zfloat = struct
  open Pervasives
  type t = float
  let ( ++ )    = ( +. )
  and ( -- )    = ( -. )
  and ( ** )    = ( *. )
  and ( // )    = fun x y -> floor(x/.y)
  and (<<= )    = (<= )
  and ( == )    = ( = )
  and zro      = 0.
  and un        = 1.
  and deux      = 2.
  and trois     = 3.
  and six       = 6.
  and douze     = 12.
  and of_string = float_of_string
  and string_of x = Printf.sprintf "%0.0f" x
  and cubroot x = floor(x**(1.0/.3.0))
end

(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                        Surface d'un triangle                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

module Surface:Surf = functor(Z:Entiers) -> struct
  open Z

  (* a,b  = coefficients du triangle.               *)
  (* q    = b/a                                     *)
  (* next = triangle suivant s'il a t dj visit *)
  type coef = {a:t; b:t; q:t; next:coef Lazy.t}


  (* enregistre a,b *)
  let rec make(u,v) =
    let (a,b) = if u <<= v then (u,v) else (v,u) in
    let rec c = {a=u; b=v; q=v//u; next=lazy(prolonge c)} in c

  and prolonge ({a=a; b=b; q=q} as coef) = let r = b -- q**a in make(r,a)

  (* calcul rcursif de surface *)
  let rec surf {a=a; b=b; q=q; next=next} n s =
    if b <<= n then                             (* hauteur >= 1 *)
      let nb = n//b and qnb = (n**q)//b in
      surf (Lazy.force next)
           (n -- a**(qnb ++ un))
           (s ++ ((nb ++ un)**(deux**qnb -- q**nb ++ deux))//deux)
    else if zro <<= n then s ++ (n // a ++ un) (* hauteur < 1 *)
    else s

  let surface coef n = surf coef n zro

end (* Surface *)

(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                        Volume d'un ttradre                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

module Volume:Vol = functor (Z:Entiers) -> struct

  module S = Surface(Z)
  open Z

  (* a,b,c = coefficients du ttradre        *)
  (* p     = profondeur maximale              *)
  (* h     = hauteur additions brutales       *)
  (* next  = dtail des calculs selon les cas *)
  type coef  = {a:t; b:t; c:t; h:t; p:int; next:calcul Lazy.t}
  and calcul =
    | Clos  of t                          (* formule close             *)
    | Sym   of coef * S.coef              (* a=b, 1/2-ttr. et tranche *)
    | Asym1 of t * coef * coef * S.coef   (* cas emboit               *)
    | Asym2 of t * coef * coef * S.coef   (* cas crois                *)
           (* q    tt+   tt-   face(a,b) *)

  (* enregistre les coefficients *)
  let rec make(u,v,w,h,p) =
    let (u,v)   = if v<<=u then (v,u)   else (u,v) in
    let (a,b,c) = if v<<=w then (u,v,w) else if u<<=w then (u,w,v) else (w,u,v) in
    let rec coef = {a=a; b=b; c=c; h=h; p=p; next=lazy(prolonge coef)} in coef

  and prolonge ({a=a; b=b; c=c; h=h; p=p} as coef) =
    if a == b then
      let q = c//a in if c == q**a
        then Clos(q)
        else Sym(make(a,deux**a,c,h,p), S.make(deux**a,c))

    else let q = (c++b--un)//b in if q**a <<= c
    then Asym2(q, make(a,b--a,c--q**a,h,p-1), make(b--a,b,q**b--c,h,p-1), S.make(a,b))

    else let q = q--un in
    Asym1(q, make(a,b--a,c--q**a,h,p-1), make(b--a,b,c--q**b,h,p-1), S.make(a,b))

  (* formule close V(1,1,q,N) *)
  let v(q,n) =
    let r   = n//q  in
    let num = r++un in
    let den = r**(deux**r++un)**q**q -- trois**r**q**(deux**n++trois)
                                ++ six**(n++un)**(n++deux)
    in (num ** den)//douze

  (* addition des surfaces horizontales *)
  let rec addition coef c n v =
    if zro <<= n then addition coef c (n--c) (v ++ S.surface coef n)
    else v

  (* calcul rcursif du volume, addition de surfaces si profondeur = 0 ou hauteur <= hmax *)
  let rec volume ({a=a; b=b; c=c; h=h; p=p; next=next} as coef) n =

    if b <<= n then match Lazy.force next with (* hauteur >= 1 *)
      | Clos(q)  -> v(q,n//a)
      | Sym(c,t) -> deux**(volume c (n--a)) ++ (S.surface t n)

      | Asym1(q,g,d,t) -> if h**c <<= n & p <> 0
        then let qnc = (q**n)//c in
             v(q,qnc) ++ (volume g (n -- a**qnc --a))
                      -- (volume d (n ++ a -- b**(qnc++deux)))
        else addition t c n zro

      | Asym2(q,g,d,t) -> if h**c <<= n & p <> 0
        then let qnc = (q**n)//c in
             v(q,qnc) ++ (volume g (n -- a**qnc -- a))
                      -- (volume d (b**qnc -- n -- un))
        else addition t c n zro

    else if zro <<= n then (n//a)++un else zro    (* h < 1 *)

end (* Volume *)


(*+--------------------------------------------------------------------------+
  |                                                                          |
  |               Recherche d'un entier de Hamming ou d'un rang              |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

module Hamming:Ham = functor (Z:Entiers) -> struct

  module S = Surface(Z)
  module V = Volume(Z)
  open Z
  open S
  open V

  (* vrifie un volume *)
  let check coef n v =

    message(fun _ -> sprintf "contrle %s -> " (string_of n));
    let k = volume coef n in
    mesg(fun _ -> sprintf "%s\n" (string_of k));
    if not(k==v) then failwith "contrle ngatif"


  (* tri selon l'ordre de Hamming *)
  let cmp a b c (x,y,z) (x',y',z') =
    let d = a**(x--x') ++ b**(y--y') ++ c**(z--z') in
         if d ++ y ++ z <<= zro   then true
    else if zro <<= d -- y' -- z' then false
    else failwith "comparaison douteuse"

  let nth a b c l n =
    let rec loop n l = if n == zro then List.hd l else loop (n--un) (List.tl l) in
    loop n (Sort.list (cmp a b c) l)

                  (*+--------------------------------------+
                    |  Encadre le k-me nombre de Hamming  |
                    +--------------------------------------+*)

  let recherche hmax pmax a b c k =

    let deuxabc = deux ** a ** b ** c  in
    let sixabc  = trois ** deuxabc     in
    let coef = V.make(a,b,c,hmax,pmax) in

    (* applique la mthode de Newton tant que l'encadrement se resserre *)
    (* hypothse : nmin <= n < nmax, vmin <= k < vmax *)
    let rec newton(nmin,vmin,nmax,vmax,n,k) =
      if vmax -- vmin <<= douze then (nmin,vmin,nmax,vmax)
      else begin

        message(fun _ -> sprintf "Newton   %s -> " (string_of n));
        let v = volume coef n in
        mesg(fun _ -> sprintf "%s\n" (string_of v));
        if v == k then (n,v,nmax,vmax)
        else let n' = n -- (deuxabc**(v--k))//(n**n) in if v <<= k

          then if nmax <<= n' or v -- vmin <<= douze or vmax -- v <<= douze
          then linaire(n,v,nmax,vmax,k)
          else newton(n,v,nmax,vmax,n',k)
             
          else if n' <<= nmin or v -- vmin <<= douze or vmax -- v <<= douze
          then linaire(nmin,vmin,n,v,k) 
          else newton(nmin,vmin,n,v,n',k)

      end

    (* termine la recherche par interpolation linaire *)
    and linaire(nmin,vmin,nmax,vmax,k) =
      if vmax -- vmin <<= douze then (nmin,vmin,nmax,vmax)
      else begin

        let n = nmax -- ((vmax--k)**(nmax--nmin))//(vmax--vmin) in
        message(fun _ -> sprintf "linaire %s -> " (string_of n));
        let v = volume coef n in
        mesg(fun _ -> sprintf "%s\n" (string_of v));
             if v == k  then (n,v,nmax,vmax)
        else if v <<= k then linaire(n,v,nmax,vmax,k)
        else                 linaire(nmin,vmin,n,v,k)
      end
    in
     
    let n0 = cubroot(k**sixabc) in
    let (p,u,q,v) = newton(zro,zro,n0++un,k++a++b++c,n0,k) in
    if v -- u <<= douze then (p,u,q,v)
    else let (_,_,q,v) = linaire(p,u,q,v,k++deux) in (p,u,q,v)


         (* +--------------------------------------------------+
            |  Liste tous les points compris entre deux plans  |
            +--------------------------------------------------+ *)

  let liste a b c p q d =

    message(fun _ -> "balayage ");
    let surf = surface (S.make(a,b)) in

    (* empile d points  y et z constants *)
    let rec ligne x y z d l =
      mesg(fun _ -> ".");
      if d == un then (x,y,z)::l else ligne (x--un) y z (d--un) ((x,y,z)::l) in

    (* empile d points  z constant, origine en (x,y) *)
    let rec triangle x y z p q d l =
      if not(b <<= q) then ligne (x++q//a) y z d l
      else
        let t = p//(deux**b) in
        let r = b**(t++un)   in
        let e = surf(q--r) -- surf(p--r) in
        let (d,l) = if e == zro then (d,l)
                    else (d--e,triangle x (y++t++un) z (p--r) (q--r) e l) in
        if d == zro then l
        else 
          let t = p//(deux**a) in
          let r = a**(t++un)   in
          triangle (x++t++un) y z (p--r) (q--r) d l in

    (* empile d points  partir de la cote z *)
    let rec ttradre z p q d l =
      let e = surf(q) -- surf(p) in
      let (d,l) = if e == zro then (d,l)
                  else (d--e,triangle zro zro z p q e l) in
      if d == zro then l
      else ttradre (z++un) (p--c) (q--c) d l in

    let l = ttradre zro p q d [] in mesg(fun _ -> "\n"); l

                  (*+--------------------------------------+
                    |  Calcule le k-me nombre de Hamming  |
                    +--------------------------------------+*)

  let hamming hmax pmax a b c k =

    let (p,u,q,v) = recherche hmax pmax a b c k in
    let coef = V.make(a,b++un,c++un,hmax,pmax)  in
    check coef p u;
    check coef q v;
    let l = liste a b c p q (v--u)   in
    let (x,y,z) = nth a b c l (k--u) in
    message(fun _ -> "termin\n\n");
    printf "x=%s y=%s z=%s\n" (string_of x) (string_of y) (string_of z)


                (*+------------------------------------------+
                  |  Calcule le rang d'un nombre de Hamming  |
                  +------------------------------------------+*)

  let rang hmax pmax a b c x y z =

    let n = x**a ++ y**b ++ z**c in
    let epsilon = n//b ++ un     in
    let nmin = n -- epsilon and nmax = n ++ epsilon in
    let coef  = V.make(a,b,c,hmax,pmax)
    and coef' = V.make(a,b++un,c++un,hmax,pmax) in

    (* calcule les volumes suprieur et infrieur *)
    let vol(n) =
      message(fun _ -> sprintf "plan sup %s -> " (string_of n));
      let k = volume coef n in
      mesg(fun _ -> sprintf "%s\n" (string_of k));
      check coef' n k;
      k
    in
    let rmin = vol(nmin) and rmax = vol(nmax) in
    message(fun _ -> "termin\n\n");
    if rmax == rmin ++ un then printf "r=%s\n" (string_of rmin)
    else printf "%s <= r < %s\n" (string_of rmin) (string_of rmax)

end (* Hamming *)

(* +------------------------------------------------------------------------+
   |                                                                        |
   |                          Interface ligne de commande                   |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

module Frontend(Z:Entiers) = struct

  module H = Hamming(Z)
  open Z
  open H
  open String


  (* Valeurs approches de ln(2), ln(3), ln(5). Avec 50 chiffres on a de quoi voir venir *)
  let ln2 = "10000000000000000000000000000000000000000000000000"
  and ln3 = "15849625007211561814537389439478165087598144076924"
  and ln5 = "23219280948873623478703194294893901758648313930245"

  (* dcode un nombre *)
  let int s =
    let l = length(s) in
    if l = 0 then zro
    else let (e,i) = match s.[l-1] with
      |('0'..'9') -> 0, l
      | 'K' -> 3,  l-1
      | 'M' -> 6,  l-1
      | 'G' -> 9,  l-1
      | 'T' -> 12, l-1
      | 'P' -> 15, l-1
      | 'E' -> 18, l-1
      | 'Z' -> 21, l-1
      | 'Y' -> 24, l-1
      |_    -> failwith "suffixe inconnu"
    in of_string ((sub s 0 i) ^ (make e '0'))

  let main hmax pmax x y z =
  (* si z est dfini on veut un rang. Sinon on cherche le y-me nb   *)
  (* de Hamming. Bidouillage (infme ?) pour avoir le -n facultatif. *)
    
    let rg = z <> "" in
    let hmax = int(hmax) and x = int(x) and y = int(y) and z = int(z) in
    let l = length((string_of x) ^ (string_of y) ^ (string_of z)) + 4 in
    let a = of_string(sub ln2 0 l)
    and b = of_string(sub ln3 0 l)
    and c = of_string(sub ln5 0 l) in

    if rg then rang    hmax pmax a b c x y z
          else hamming hmax pmax a b c y
    
end (* Frontend *)

(* lance les calculs *)
open Arg

let x     = ref "" and y = ref "" and z = ref ""
and float = ref false
and hmax  = ref "1000"
and pmax  = ref(5)

let helpmsg = "usage : " ^ Sys.argv.(0) ^ " (-r x y z | -n n) [-h hmax] [-p prof] [-debug] ( -float | -num)]
Par dfaut : -n, -h 1000, -p 5.
"
let _ =
  parse [ ("-r",        String(fun s -> x := s),    "    cherche un rang");
          ("-n",        String(fun s -> y := s),    "    cherche un nombre");
          ("-h",        String(fun s -> hmax := s), "    hauteur seuil");
          ("-p",        Int   (fun p -> pmax := p), "    profondeur maxi");
          ("-debug",    Set(debug),                 "    dbogage");
          ("-float",    Set(float),                 "    entier = flottant");
          ("-num",      Clear(float),               "    entier = num") ]
          (fun s -> if !y = "" then y := s else z := s)
          helpmsg

module Fnum   = Frontend(Znum)
module Ffloat = Frontend(Zfloat)

let _ =
  if !y <> ""
  then (if !float then Ffloat.main else Fnum.main) !hmax !pmax !x !y !z
  else prerr_endline helpmsg
